"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const node_sql_parser_1 = require("@joe-re/node-sql-parser");
const log4js = require("log4js");
const vscode_languageserver_1 = require("vscode-languageserver");
const logger = log4js.getLogger();
const CLAUSES = [
    { label: 'WHERE', kind: vscode_languageserver_1.CompletionItemKind.Text },
    { label: 'ORDER BY', kind: vscode_languageserver_1.CompletionItemKind.Text },
    { label: 'GROUP BY', kind: vscode_languageserver_1.CompletionItemKind.Text },
    { label: 'LIMIT', kind: vscode_languageserver_1.CompletionItemKind.Text }
];
function extractExpectedLiterals(expected) {
    return expected.filter(v => v.type === 'literal')
        .map(v => v.text)
        .filter((v, i, self) => self.indexOf(v) === i)
        .map(v => ({ label: v }));
}
function getLastToken(sql) {
    const match = sql.match(/^(?:.|\s)*[^A-z0-9](.*?)$/);
    if (!match) {
        return sql;
    }
    return match[1];
}
function getColumnRefByPos(columns, pos) {
    return columns.find(v => (v.location.start.line === pos.line + 1 && v.location.start.column <= pos.column) &&
        (v.location.end.line === pos.line + 1 && v.location.end.column >= pos.column));
}
function getFromNodeByPos(fromNodes, pos) {
    return fromNodes.find(v => (v.location.start.line === pos.line + 1 && v.location.start.column <= pos.column) &&
        (v.location.end.line === pos.line + 1 && v.location.end.column >= pos.column));
}
function toCompletionItemFromTable(table) {
    return {
        label: table.tableName,
        detail: `table ${table.tableName}`,
        kind: vscode_languageserver_1.CompletionItemKind.Text
    };
}
function toCompletionItemFromColumn(column) {
    return {
        label: column.columnName,
        detail: `column ${column.description}`,
        kind: vscode_languageserver_1.CompletionItemKind.Text
    };
}
function getCandidatesFromColumnRefNode(columnRefNode, schema) {
    const tableCandidates = schema.filter(v => v.tableName.startsWith(columnRefNode.table)).map(v => toCompletionItemFromTable(v));
    const columnCandidates = Array.prototype.concat.apply([], schema.filter(v => tableCandidates.map(v => v.label).includes(v.tableName)).map(v => v.columns)).map((v) => toCompletionItemFromColumn(v));
    return tableCandidates.concat(columnCandidates);
}
function isCursorOnFromClause(sql, pos) {
    try {
        const ast = node_sql_parser_1.Parser.parse(sql);
        return !!getFromNodeByPos(ast.from || [], pos);
    }
    catch (_e) {
        return false;
    }
}
function getCandidatedFromIncompleteSubquery(params) {
    let candidates = [];
    const { schema, incompleteSubquery, pos } = params;
    const parsedFromClause = getFromNodesFromClause(incompleteSubquery.text);
    try {
        node_sql_parser_1.Parser.parse(incompleteSubquery.text);
    }
    catch (e) {
        if (e.name !== 'SyntaxError') {
            throw e;
        }
        const fromText = incompleteSubquery.text;
        const newPos = parsedFromClause ? {
            line: pos.line - (incompleteSubquery.location.start.line - 1),
            column: pos.column - incompleteSubquery.location.start.column + 1
        } : { line: 0, column: 0 };
        candidates = complete(fromText, newPos, schema).candidates;
    }
    return candidates;
}
function createTablesFromFromNodes(fromNodes) {
    return fromNodes.reduce((p, c) => {
        if (c.type !== 'subquery') {
            return p;
        }
        const columns = c.subquery.columns.map(v => {
            if (typeof v === 'string') {
                return null;
            }
            return { columnName: v.as || v.expr.column || '', description: 'alias' };
        });
        return p.concat({ database: null, columns, tableName: c.as });
    }, []);
}
function getCandidatesFromError(target, schema, pos, e, fromNodes) {
    let candidates = extractExpectedLiterals(e.expected);
    const candidatesLiterals = candidates.map(v => v.label);
    if (candidatesLiterals.includes("'") || candidatesLiterals.includes('"')) {
        return [];
    }
    if (candidatesLiterals.includes('.')) {
        candidates = candidates.concat(schema.map(v => toCompletionItemFromTable(v)));
    }
    const lastChar = target[target.length - 1];
    logger.debug(`lastChar: ${lastChar}`);
    if (lastChar === '.') {
        const removedLastDotTarget = target.slice(0, target.length - 1);
        if (isCursorOnFromClause(removedLastDotTarget, { line: pos.line, column: pos.column - 1 })) {
            return [];
        }
        const tableName = getLastToken(removedLastDotTarget);
        const subqueryTables = createTablesFromFromNodes(fromNodes);
        const attachedAlias = schema.concat(subqueryTables).map(v => {
            const as = fromNodes.filter((v2) => v.tableName === v2.table).map(v => v.as);
            return Object.assign({}, v, { as: as ? as : [] });
        });
        let table = attachedAlias.find(v => v.tableName === tableName || v.as.includes(tableName));
        if (table) {
            candidates = table.columns.map(v => toCompletionItemFromColumn(v));
        }
    }
    return candidates;
}
function getFromNodesFromClause(sql) {
    try {
        return node_sql_parser_1.Parser.parseFromClause(sql);
    }
    catch (_e) {
        // no-op
        return null;
    }
}
function getRidOfAfterCursorString(sql, pos) {
    return sql.split('\n').filter((_v, idx) => pos.line >= idx).map((v, idx) => idx === pos.line ? v.slice(0, pos.column) : v).join('\n');
}
function complete(sql, pos, schema = []) {
    logger.debug(`complete: ${sql}, ${JSON.stringify(pos)}`);
    let candidates = [];
    let error = null;
    const target = getRidOfAfterCursorString(sql, pos);
    logger.debug(`target: ${target}`);
    try {
        candidates = CLAUSES.concat([]);
        const ast = node_sql_parser_1.Parser.parse(target);
        const ar = new node_sql_parser_1.AstReader(ast);
        logger.debug(`ast: ${JSON.stringify(ar.getAst())}`);
        if (!ar.getAst().distinct) {
            candidates.push({ label: 'DISTINCT', kind: vscode_languageserver_1.CompletionItemKind.Text });
        }
        if (Array.isArray(ar.getAst().columns)) {
            const selectColumnRefs = ar.getAst().columns.map((v) => v.expr).filter((v) => !!v);
            const whereColumnRefs = ar.getAst().where || [];
            const columnRef = getColumnRefByPos(selectColumnRefs.concat(whereColumnRefs), pos);
            logger.debug(JSON.stringify(columnRef));
            if (columnRef) {
                candidates = candidates.concat(getCandidatesFromColumnRefNode(columnRef, schema));
            }
        }
        if (Array.isArray(ar.getAst().from)) {
            const fromTable = getFromNodeByPos(ar.getAst().from || [], pos);
            if (fromTable && fromTable.type === 'table') {
                candidates = candidates.concat(schema.map(v => toCompletionItemFromTable(v)))
                    .concat([{ label: 'INNER JOIN' }, { label: 'LEFT JOIN' }]);
                if (fromTable.join && !fromTable.on) {
                    candidates.push({ label: 'ON' });
                }
            }
        }
    }
    catch (e) {
        logger.debug('error');
        logger.debug(e);
        if (e.name !== 'SyntaxError') {
            throw e;
        }
        const parsedFromClause = getFromNodesFromClause(sql);
        const fromNodes = parsedFromClause && parsedFromClause.from || [];
        const fromNodeOnCursor = getFromNodeByPos(fromNodes || [], pos);
        if (fromNodeOnCursor && fromNodeOnCursor.type === 'incomplete_subquery') {
            candidates = getCandidatedFromIncompleteSubquery({
                sql,
                pos,
                incompleteSubquery: fromNodeOnCursor,
                schema
            });
        }
        else {
            candidates = getCandidatesFromError(target, schema, pos, e, fromNodes);
        }
        error = { label: e.name, detail: e.message, line: e.line, offset: e.offset };
    }
    const lastToken = getLastToken(target);
    logger.debug(`lastToken: ${lastToken}`);
    console.log(JSON.stringify(candidates));
    candidates = candidates.filter(v => v.label.startsWith(lastToken));
    return { candidates, error };
}
exports.default = complete;
//# sourceMappingURL=complete.js.map